# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""


class Model(object):
    def __init__(self):
        pass

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        pass

    def normalize_parameters(self):
        pass

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        pass
